#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019 SKALE Labs
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Get SKALE validator data """
from web3 import Web3
from skale.contracts import BaseContract


class ValidatorsData(BaseContract):
    def get_reward_period(self):
        constants = self.skale.get_contract_by_name('constants')
        return constants.contract.functions.rewardPeriod().call()

    def get_delta_period(self):
        constants = self.skale.get_contract_by_name('constants')
        return constants.contract.functions.deltaPeriod().call()

    def get_validated_array(self, node_id, account):
        node_id_bytes = Web3.sha3(node_id)
        return self.contract.functions.getValidatedArray(node_id_bytes).call(
            {'from': account})



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019 SKALE Labs
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Get SKALE chain data """

from Crypto.Hash import keccak

from skale.contracts import BaseContract
from skale.utils.helper import format, ip_from_bytes, public_key_to_address

from skale.dataclasses.current_node_info import CurrentNodeInfo
from skale.dataclasses.schain_node_info import SchainNodeInfo

FIELDS = [
    'name', 'owner', 'indexInOwnerList', 'partOfNode', 'lifetime', 'startDate',
    'deposit', 'index'
]

PORTS_PER_SCHAIN = 11
NUMBER_OF_PORTS = 3


class SChainsData(BaseContract):
    def __get_raw(self, name):
        return self.contract.functions.schains(name).call()

    @format(FIELDS)
    def get(self, id):
        return self.__get_raw(id)

    @format(FIELDS)
    def get_by_name(self, name):
        id = self.name_to_id(name)
        return self.__get_raw(id)

    def get_schains_for_owner(self, account):
        schains = []
        list_size = self.get_schain_list_size(account)

        for i in range(0, list_size):
            id = self.get_schain_id_by_index_for_owner(account, i)
            schain = self.get(id)
            schains.append(schain)
        return schains

    def get_schain_list_size(self, account):
        return self.contract.functions.getSchainListSize(account).call(
            {'from': account})

    def get_schain_id_by_index_for_owner(self, account, index):
        return self.contract.functions.schainIndexes(account, index).call()

    def get_current_node_for_schain_config(self, schain_name: str, node_id: int) -> CurrentNodeInfo:
        node = self.skale.nodes_data.get(node_id)
        schain_base_port = self.get_schain_base_port_on_node(schain_name, node_id, node['port'])
        return CurrentNodeInfo(
            node_name=node['name'],
            node_id=node_id,
            base_port=schain_base_port,
            bind_ip=ip_from_bytes(node['ip'])
        ).to_config()

    def get_nodes_for_schain_config(self, name):
        nodes_info = []
        nodes = self.get_nodes_for_schain(name)

        for i, node in enumerate(nodes, 1):
            pk = node['publicKey'].hex()
            base_port = self.get_schain_base_port_on_node(name, node['id'], node['port'])

            node_info = SchainNodeInfo(
                node_name=node['name'],
                node_id=node['id'],
                base_port=base_port,

                schain_index=i,
                ip=ip_from_bytes(node['ip']),
                public_key=pk,
                public_ip=ip_from_bytes(node['publicIP']),
                owner=public_key_to_address(pk)
            ).to_config()
            nodes_info.append(node_info)
        return nodes_info

    def get_schain_base_port_on_node(self, schain_name, node_id, node_port):
        schains = self.get_schains_for_node(node_id)
        schain_index = self.get_schain_index_in_node(schain_name, schains)
        return self.calc_schain_base_port(node_port, schain_index)

    def get_schain_index_in_node(self, schain_name, node_schains):
        for index, schain in enumerate(node_schains):
            if schain_name == schain['name']:
                return index
        return -1

    def calc_schain_base_port(self, node_base_port, schain_index):
        return node_base_port + schain_index * PORTS_PER_SCHAIN

    def get_nodes_for_schain(self, name):
        nodes = []
        ids = self.get_node_ids_for_schain(name)
        for id in ids:
            node = self.skale.nodes_data.get(id)
            node['id'] = id
            nodes.append(node)
        return nodes

    def get_node_ids_for_schain(self, name):
        id = self.name_to_id(name)
        return self.contract.functions.getNodesInGroup(id).call()

    def get_schain_ids_for_node(self, node_id):
        return self.contract.functions.getSchainIdsForNode(node_id).call()

    def get_schains_for_node(self, node_id):
        schains = []
        schain_ids = self.get_schain_ids_for_node(node_id)
        for schain_id in schain_ids:
            schain = self.get(schain_id)
            schain['active'] = True if self.schain_active(schain) else False
            schains.append(schain)
        return schains

    def name_to_id(self, name):
        keccak_hash = keccak.new(data=name.encode("utf8"), digest_bits=256)
        return keccak_hash.hexdigest()

    def get_all_schains_ids(self):
        return self.contract.functions.getSchains().call()

    def get_schains_number(self):
        return self.contract.functions.numberOfSchains().call()

    def schain_active(self, schain):
        if schain['name'] != '' and \
                schain['owner'] != '0x0000000000000000000000000000000000000000':
            return True



import random
from skale.utils.random_names.vocabulary import CONSTELLATIONS, STARS, \
    SOUND_AND_APPEARANCE_ADJECTIVES, POSITIVE_AND_TIME_ADJECTIVES


def generate_random_node_name():
    return generate_name(CONSTELLATIONS, POSITIVE_AND_TIME_ADJECTIVES)


def generate_random_schain_name():
    return generate_name(STARS, SOUND_AND_APPEARANCE_ADJECTIVES)


def generate_name(noun_dict, adjective_dict):
    noun = get_random_word_from_dict(noun_dict)
    adjective = get_random_word_from_dict(adjective_dict)
    return f'{adjective}-{noun}'


def get_random_word_from_dict(vocabulary_dict):
    return random.choice(vocabulary_dict).lower().replace(" ", "-")


if __name__ == "__main__":
    node_name = generate_random_node_name()
    schain_name = generate_random_schain_name()
    print(f'Node name: {node_name}, sChain name: {schain_name}')



from skale.dataclasses.node_info import NodeInfo


class SchainNodeInfo(NodeInfo):
    def __init__(self, node_id, node_name, base_port, public_key, owner, schain_index, ip,
                 public_ip):
        self.public_key = public_key
        self.owner = owner
        self.schain_index = schain_index
        self.ip = ip
        self.public_ip = public_ip
        super().__init__(node_id, node_name, base_port)

    def to_config(self):
        return {**super().to_config(), **{
            'publicKey': self.public_key,
            'owner': self.owner,
            'schainIndex': self.schain_index,
            'ip': self.ip,
            'publicIP': self.public_ip
        }}



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019 SKALE Labs
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" SKALE validator functions """

from skale.contracts import BaseContract


class Validators(BaseContract):
    pass



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019 SKALE Labs
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Get SKALE node data """

import socket

from Crypto.Hash import keccak
from web3.exceptions import BadFunctionCallOutput

from skale.contracts import BaseContract
from skale.utils.helper import format

FIELDS = [
    'name', 'ip', 'publicIP', 'port', 'publicKey', 'start_date',
    'leaving_date', 'last_reward_date', 'second_address'
]
COMPACT_FIELDS = ['schainIndex', 'nodeID', 'ip', 'basePort']
SCHAIN_CONFIG_FIELDS = [
    'schainIndex', 'nodeID', 'nodeName', 'ip', 'basePort',
    'publicKey', 'publicIP', 'owner',
    'httpRpcPort', 'httpsRpcPort', 'wsRpcPort', 'wssRpcPort'
]


class NodesData(BaseContract):
    def __get_raw(self, node_id):
        try:
            return self.contract.functions.nodes(node_id).call()
        except (ValueError, BadFunctionCallOutput):
            return None

    @format(FIELDS)
    def get(self, node_id):
        return self.__get_raw(node_id)

    @format(FIELDS)
    def get_by_name(self, name):
        name_hash = self.name_to_id(name)
        id = self.contract.functions.nodesNameToIndex(name_hash).call()
        return self.__get_raw(id)

    def get_active_node_ids(self):
        return self.contract.functions.getActiveNodeIds().call()

    def get_active_node_ips(self):
        return self.contract.functions.getActiveNodeIPs().call()

    def get_active_node_ids_by_address(self, account):
        return self.contract.functions.getActiveNodesByAddress().call(
            {'from': account})

    def name_to_id(self, name):
        keccak_hash = keccak.new(data=name.encode("utf8"), digest_bits=256)
        return keccak_hash.hexdigest()

    def is_node_name_available(self, name):
        node_id = self.name_to_id(name)
        return not self.contract.functions.nodesNameCheck(node_id).call()

    def is_node_ip_available(self, ip):
        ip_bytes = socket.inet_aton(ip)
        return not self.contract.functions.nodesIPCheck(ip_bytes).call()

    def node_name_to_index(self, name):
        name_hash = self.name_to_id(name)
        return self.contract.functions.nodesNameToIndex(name_hash).call()



import logging

from web3 import Web3, WebsocketProvider, HTTPProvider
from urllib.parse import urlparse

import skale.contracts as contracts
from skale.contracts_info import CONTRACTS_INFO
from skale.utils.helper import get_abi

logger = logging.getLogger(__name__)


class Skale:
    def __init__(self, endpoint, abi_filepath):
        logger.info(f'Init skale-py, connecting to {endpoint}')
        provider = self.get_provider(endpoint)
        self.web3 = Web3(provider)
        self.abi = get_abi(abi_filepath)
        self.__contracts = {}
        self.__contracts_info = {}
        self.nonces = {}

        self.__init_contracts_info()
        self.__init_contracts()

    def get_provider(self, endpoint):
        scheme = urlparse(endpoint).scheme
        if scheme == 'ws' or scheme == 'wss':
            return WebsocketProvider(endpoint)
        if scheme == 'http' or scheme == 'https':
            return HTTPProvider(endpoint)
        raise Exception(
            'Wrong endpoint option.'
            'Supported endpoint schemes: http/https/ws/wss'
        )

    def __init_contracts_info(self):
        for contract_info in CONTRACTS_INFO:
            self.__add_contract_info(contract_info)

    def __add_contract_info(self, contract_info):
        self.__contracts_info[contract_info.name] = contract_info

    def __init_contracts(self):
        self.add_lib_contract('contract_manager', contracts.ContractManager)
        for name in self.__contracts_info:
            info = self.__contracts_info[name]
            if info.upgradeable:
                self.init_upgradeable_contract(info)
            else:
                self.add_lib_contract(info.name, info.contract_class)

    def init_upgradeable_contract(self, contract_info):
        address = self.get_contract_address(contract_info.contract_name)
        self.add_lib_contract(contract_info.name, contract_info.contract_class,
                              address)

    def add_lib_contract(self, name, contract_class, contract_address=None):
        address = contract_address or self.get_contract_address_by_name(
            self.abi, name)
        logger.info(f'Initialized: {name} at {address}')
        abi = self.get_contract_abi_by_name(self.abi, name)
        self.add_contract(name, contract_class(self, name, address, abi))

    def get_contract_address_by_name(self, abi, name):
        return abi.get(f'skale_{name}_address') or abi.get(f'{name}_address')

    def get_contract_abi_by_name(self, abi, name):  # todo: unify abi key names
        return abi.get(f'skale_{name}_abi') or abi.get(
            f'{name}_abi') or abi.get(f'{name}_functionality_abi') or abi.get(
            f'{name}_data_abi')

    def add_contract(self, name, skale_contract):
        logger.debug(f'Init contract: {name}')
        self.__contracts[name] = skale_contract

    def get_contract_address(self, name):
        return self.contract_manager.get_contract_address(name)

    def get_contract_by_name(self, name):
        return self.__contracts[name]

    def __getattr__(self, name):
        if name not in self.__contracts:
            raise AttributeError(name)
        return self.get_contract_by_name(name)



# flake8: noqa: E402

import sys

if sys.version_info < (3, 5):
    raise EnvironmentError("Python 3.5 or above is required")

from skale.main import Skale
from skale.contracts import (Manager, Token, Nodes, SChains,
                             Validators, Groups, BaseContract, Constants)



from enum import Enum


class SkaledPorts(Enum):
    PROPOSAL = 0
    CATCHUP = 1
    HTTP_JSON = 2
    WS_JSON = 3
    BINARY_CONSENSUS = 4
    ZMQ_BROADCAST = 5
    MTA = 6
    HTTPS_JSON = 7
    WSS_JSON = 8



from skale.skaled_ports import SkaledPorts


class NodeInfo():
    def __init__(self, node_id, node_name, base_port):
        self.node_id = node_id
        self.node_name = node_name
        self.base_port = base_port
        self.calc_ports()

    def calc_ports(self):
        self.http_rpc_port = self.base_port + SkaledPorts.HTTP_JSON.value
        self.https_rpc_port = self.base_port + SkaledPorts.HTTPS_JSON.value
        self.ws_rpc_port = self.base_port + SkaledPorts.WS_JSON.value
        self.wss_rpc_port = self.base_port + SkaledPorts.WSS_JSON.value

    def to_config(self):
        return {
            'nodeID': self.node_id,
            'nodeName': self.node_name,
            'basePort': self.base_port,
            'httpRpcPort': self.http_rpc_port,
            'httpsRpcPort': self.https_rpc_port,
            'wsRpcPort': self.ws_rpc_port,
            'wssRpcPort': self.wss_rpc_port
        }
